export function string__to_data(value: string): string {
  return value;
}

// eslint-disable-next-line @typescript-eslint/no-explicit-any
export function string__from_data(data: any): string {
  if (typeof data !== "string") {
    {{RAISE_DESERIALIZATION_ERROR}};
  }
  return data;
}