export function bool__to_data(value: boolean): boolean {
  return value;
}

// eslint-disable-next-line @typescript-eslint/no-explicit-any
export function bool__from_data(data: any): boolean {
  if (typeof data !== "boolean") {
    {{RAISE_DESERIALIZATION_ERROR}};
  }
  return data;
}